/*
 * $Id: dfilter.c 24238 2008-01-31 23:10:09Z stig $
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 2001 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <string.h>
/* PMC Wildcard Support	-  start */
#include "globals.h"
/* PMC Wildcard Support	-  end */


#include "dfilter-int.h"
#include "syntax-tree.h"
#include "gencode.h"
#include "semcheck.h"
#include "dfvm.h"
#include <epan/epan_dissect.h>
#include "dfilter.h"
#include "dfilter-macro.h"
#include <epan/report_err.h>

#define DFILTER_TOKEN_ID_OFFSET	1

/* Global error message space for dfilter_compile errors */
static gchar dfilter_error_msg_buf[1024];
const gchar *dfilter_error_msg;	/* NULL when no error resulted */



/* PMC Wilcard Support	- start */
static HMODULE wiresharkLib = NULL;
/* PMC Wilcard Support	- end */

/* From scanner.c */
void df_scanner_text(const char *text);
void    df_scanner_cleanup(void);
int     df_lex(void);

/* Holds the singular instance of our Lemon parser object */
static void*	ParserObj = NULL;

void
dfilter_fail(const char *format, ...)
{
	va_list	args;

	/* If we've already reported one error, don't overwite it */
	if (dfilter_error_msg != NULL)
		return;

	va_start(args, format);

	g_vsnprintf(dfilter_error_msg_buf, sizeof(dfilter_error_msg_buf),
			format, args);
	dfilter_error_msg = dfilter_error_msg_buf;
	va_end(args);
}


/* Initialize the dfilter module */
void
dfilter_init(void)
{
	if (ParserObj) {
		g_message("I expected ParserObj to be NULL\n");
		/* Free the Lemon Parser object */
		DfilterFree(ParserObj, g_free);
	}
	/* Allocate an instance of our Lemon-based parser */
	ParserObj = DfilterAlloc(g_malloc);

/* Enable parser tracing by defining AM_CFLAGS
 * so that it contains "-DDFTRACE".
 */
#ifdef DFTRACE
	/* Trace parser */
	DfilterTrace(stdout, "lemon> ");
#endif

	/* Initialize the syntax-tree sub-sub-system */
	sttype_init();

	dfilter_macro_init();
}

/* Clean-up the dfilter module */
void
dfilter_cleanup(void)
{
	/* Free the Lemon Parser object */
	if (ParserObj) {
		DfilterFree(ParserObj, g_free);
	}

	/* Clean up the syntax-tree sub-sub-system */
	sttype_cleanup();
}

static dfilter_t*
dfilter_new(void)
{
	dfilter_t	*df;

	df = g_new(dfilter_t, 1);
	df->insns = NULL;
        df->deprecated = NULL;

	return df;
}

/* Given a GPtrArray of instructions (dfvm_insn_t),
 * free them. */
static void
free_insns(GPtrArray *insns)
{
	unsigned int	i;
	dfvm_insn_t	*insn;

	for (i = 0; i < insns->len; i++) {
		insn = g_ptr_array_index(insns, i);
		dfvm_insn_free(insn);
	}
	g_ptr_array_free(insns, TRUE);
}

void
dfilter_free(dfilter_t *df)
{
	int i;

	if (!df)
		return;

	if (df->insns) {
		free_insns(df->insns);
	}
	if (df->consts) {
		free_insns(df->consts);
	}

	if (df->interesting_fields) {
		g_free(df->interesting_fields);
	}

	/* clear registers */
	for (i = 0; i < df->max_registers; i++) {
		if (df->registers[i]) {
			g_list_free(df->registers[i]);
		}
	}

	g_free(df->registers);
	g_free(df->attempted_load);
	g_free(df);
}


static dfwork_t*
dfwork_new(void)
{
	dfwork_t	*dfw;

	dfw = g_new(dfwork_t, 1);

	dfw->st_root = NULL;
	dfw->syntax_error = FALSE;
	dfw->insns = NULL;
	dfw->consts = NULL;
	dfw->loaded_fields = NULL;
	dfw->interesting_fields = NULL;
	dfw->next_insn_id = 0;
	dfw->next_register = 0;
	dfw->first_constant = -1;

	return dfw;
}

static void
dfwork_free(dfwork_t *dfw)
{
	if (dfw->st_root) {
		stnode_free(dfw->st_root);
	}

	if (dfw->loaded_fields) {
		g_hash_table_destroy(dfw->loaded_fields);
	}

	if (dfw->interesting_fields) {
		g_hash_table_destroy(dfw->interesting_fields);
	}

	if (dfw->insns) {
		free_insns(dfw->insns);
	}

	if (dfw->consts) {
		free_insns(dfw->consts);
	}

	g_free(dfw);
}
/* PMC Wildcard, hello, soap Support - start */
void
left_right_trim(char *string)
{

	unsigned int i = 0;
	unsigned int j;

	if( NULL == string )
		return;

	/* Trim spaces, tabs, left and right paranthesis from end */
	while( (string [i] == ' ') || (string[i] == '\t') )
		i++;

	if(i > 0)
	{
		for(j = 0; j < strlen(string); j++)
			string[j]=string[j+i];

		string[j]='\0';
	}

	/* Trim spaces, tabs, left and right paranthesis from end */
	i = strlen(string) - 1;
	while( (string[i] == ' ') || (string[i] == '\t') )
		i--;

	if(i < (strlen(string)-1))
		string[i+1] = '\0';
}

int
is_blank_line(char *line)
{
	int i = 0, len;

	if( NULL == line)
		return 0;

	len = strlen(line) - 1; /* one less for new line character */

	/* check for spaces and tabs from beginning if some
	nonspace (nontab) character is found return FALSE else return TRUE */
	for(i = 0; i < len; i++)
	{
		if( (line [i] != ' ') && (line[i] != '\t') )
			return 0;
	}
	return 1;
}


char* create_filter(char* protocol_name, char* parameter_name, char* expr)
{
	char * file_name = NULL;
	char line [ MAX_LINE_LENGTH ];			/* for storing line from file */
	char temp_line [ MAX_LINE_LENGTH ];		/* temporary variable for storing line from file */
	char * token = NULL;					/* for token return from strtok function */
	FILE *wildcard_config_file = NULL;		/* config file pointer */
	char * genarated_filter_string = NULL;	/* for filter string replacing wildcard */
	int temp_len = 0;


	if ((protocol_name == NULL)||(parameter_name == NULL))
		return NULL;

	if ( NULL == wiresharkLib )
	{
		/* loading libwireshark.dll for getting absolute path */
		wiresharkLib =  LoadLibrary(TEXT("libwireshark.dll"));
		if( NULL == wiresharkLib )
			goto CLEANUP;
	}
	full_path = (char**) GetProcAddress(wiresharkLib, "full_path_name");


	file_name = (char *) calloc ( strlen(*full_path)
		+ strlen(protocol_name)
		+ 15, sizeof (char)
		);

	if ( NULL == file_name )
		goto CLEANUP;

	sprintf( file_name, "%s\\Config\\%s.txt", *full_path, protocol_name );

	wildcard_config_file = fopen ( file_name ,"r" );

	if ( NULL == wildcard_config_file )
		goto CLEANUP;

	if ( NULL != file_name )
		free(file_name);

	memset(line, 0, sizeof(line));

	while ( fgets ( line, sizeof(line), wildcard_config_file ) != NULL )
	{
		/* checking for blank lines  */
		if( is_blank_line(line) )
			continue;

		memset(temp_line, 0, sizeof(line));
		/* copying line in temp_line except new line character  */
		strncpy( temp_line, line, strlen(line)-1  );
		/* searching protocol name from line */
		token = strtok( line, "." );
		if ( NULL == token )
			continue;
		temp_len = strlen(token);
		/* matching protocol name with token */
		if ( ( 0 == strncmp( token, protocol_name, temp_len ))
			&& ( temp_len == strlen(protocol_name) )
			)
		{
			/* searching parameter name from line */
			token = strtok( NULL, "." );
			if ( token != NULL )
			{
				// skipping the apiid
				token = strtok( NULL, END_OF_LINE_DELIMETER);
			}
			else
				continue;

			if ( token != NULL )
			{
				temp_len = strlen(token);
				/* matching parameter name with token */
				if ( ( 0 == strncmp( token, parameter_name, temp_len ) )
					&& ( temp_len == strlen(parameter_name) )
					)
				{
					/* for firat time memory allocation to genarated_filter_string */
					if ( NULL == genarated_filter_string )
					{
						if ( NULL != expr )
							/* 3 for "(()" and 1 for NULL char */
							genarated_filter_string = (char *) calloc ( strlen(temp_line)
							+ strlen(expr)
							+ 4, sizeof (char)
							);
						else
							/* 3 for "(()" and 1 for NULL char */
							genarated_filter_string = (char *) calloc ( strlen(temp_line)
							+ 4, sizeof (char)
							);
						if ( NULL == genarated_filter_string )
							goto CLEANUP;
						if ( NULL != expr )
							/* prepending "(" at the begining of genarated_filter_string */
							sprintf( genarated_filter_string, "((%s%s)",temp_line,
							expr
							);
						else
							/* prepending "(" at the begining of genarated_filter_string */
							sprintf( genarated_filter_string, "((%s)",temp_line);
					}
					else
					{
						if ( NULL != expr )
							/* 4 for "||()" and 1 for NULL char */
							genarated_filter_string = (char *) realloc(
							genarated_filter_string,
							( strlen(genarated_filter_string)
							+ strlen(temp_line)
							+ strlen(expr)
							+ 5
							) * sizeof (char)
							);
						else
							/* 4 for "||()" and 1 for NULL char */
							genarated_filter_string = (char *) realloc (
							genarated_filter_string,
							( strlen(genarated_filter_string)
							+ strlen(temp_line)
							+ 5
							) * sizeof (char) );
						if ( NULL == genarated_filter_string )
							goto CLEANUP;
						if ( NULL != expr )
							sprintf( genarated_filter_string, "%s||(%s%s)",genarated_filter_string
							,temp_line,expr);
						else
							sprintf( genarated_filter_string, "%s||(%s)",genarated_filter_string
							, temp_line );
					}
				}
			}
		}
	}
	if ( NULL != wildcard_config_file )
		fclose(wildcard_config_file);

	if ( genarated_filter_string )
	{
		genarated_filter_string = (char *) realloc ( genarated_filter_string,
										  ( strlen(genarated_filter_string) + 1 + 1 )
										  * sizeof (char)
										  );
		if ( NULL == genarated_filter_string )
			goto CLEANUP;
		sprintf( genarated_filter_string, "%s)",genarated_filter_string);
	}
	return genarated_filter_string;

CLEANUP:
	
	if ( NULL != wildcard_config_file )
		fclose(wildcard_config_file);

	if ( NULL != file_name )
		free(file_name);
	return NULL;

}



char *
create_filter_string( const char *field_string )
{
	char protocol_name[ MAX_PROTOCOL_NAME ];	/*
														** for protocol name
														** e.g. api_call
														*/
	char parameter_name[ MAX_PROTOCOL_PARAMETER_NAME ]; /*
														** for parameter name
														** e.g. TimerId
														*/
	char expr[ MAX_EXPRESSION_LENGTH ];/* for expression value e.g. == 0x00 */


	char * token = NULL;		/* for token return from strtok function */
	char * temp_token = NULL;  /* for temporary token */
	char * temp_filter_string_second = NULL; /*
										** temporary variable for storing
										** second part of filter string
										*/
	char * filter_string = NULL;	/* for filter string replacing wildcard */

	char * filter_string_apicall = NULL;/*
										** temporary variable for storing
										** api call filter string in case of "api"
										*/
	char * filter_string_api_rc = NULL;/*
										** temporary variable for storing
										** api rc filter string in case of "api"
										*/
	char * filter_string_api_mn = NULL;/*
										** temporary variable for storing
										** api mn filter string in case of "api"
										*/


	char * temp_filter_string = NULL;	/*
										** temporary variable for storing
										** filter string
										*/
	char * first_token_string = NULL;	/* filter string before wildcard(*) */
	char * second_token_string = NULL;	/* filter string after wildcard(*) */
	char * first_string_prepend = NULL;	/*
										** string to be prepend to variable
										** filter_string
										*/
	char * second_string_append = NULL;	/*
										** string to be append to variable
										** filter_string
										*/
	char * final_filter_string = NULL;	/*
										** final filter string
										** = first_string_prepend
										** + filter_string
										** + second_string_append
										*/

	int first_string_prepend_flag = 1;	/*
										** flag to check if there is string to
										** be prepend
										*/
	int second_string_append_flag = 1;	/*
										** flag to check if there is string to
										** be append
										*/
	int first_token_len = 0;	/* length of first token string	*/
	int second_token_len = 0;	/* length of second token string */
	int string_index = 0;		/* for temp array indexing */
	int expr_index = 0;			/* for expr array indexing */
	int param_index = 0;		/* for parameter name array indexing */
	int length_count;			/* length traversed for second token string */
	int expr_flag;				/* temporary flag variable for expression */
	int temp_len = 0;			/* temporary length variable */
	int protocol_string_len = 0;/* for protocol string length */
	int len =0;
	if ( NULL == field_string )
		return NULL;

	/* allocating memory for temporary filter string */
	temp_filter_string = (char *) calloc ( strlen(field_string)+1,
											sizeof (char)
										  );

	temp_filter_string_second = (char *) calloc ( strlen(field_string)+1,
											sizeof (char)
										  );

	if ( NULL == temp_filter_string || NULL == temp_filter_string_second )
		return NULL;
	/* copying original filter string in a temporary string */
	strncpy( temp_filter_string, field_string, strlen(field_string) );
	strncpy( temp_filter_string_second, field_string, strlen(field_string) );
	/* splitting temp filter string in two parts by separator as wildcard(.*.) */
	token = strstr( temp_filter_string, ".*." );

	protocol_string_len =  strlen(temp_filter_string) - strlen (token)+1;



	/* allocating memory for first token string */
	first_token_string = (char *) calloc ( protocol_string_len+1, sizeof (char) );
	if ( NULL == first_token_string )
		goto CLEANUP;
	/* copying token in first token string */
	strncpy( first_token_string, temp_filter_string, protocol_string_len );

	temp_token = temp_filter_string_second + protocol_string_len + 1;
	/* allocating memory for second token string */
	second_token_string = (char *) calloc ( strlen(temp_token)+1, sizeof (char) );
	if ( NULL == second_token_string )
		goto CLEANUP;
	/* copying token in second token string */
	strncpy( second_token_string, temp_token, strlen(temp_token) );

	free(temp_filter_string);
	temp_filter_string = NULL;
	free(temp_filter_string_second);
	temp_filter_string_second = NULL;

	/* trimming spaces and tabs from begining and end of both strings */
	left_right_trim( first_token_string );
	left_right_trim( second_token_string );

	memset(protocol_name, 0, sizeof(protocol_name));
	/* to start length and index both from index 0 */
	first_token_len = strlen(first_token_string)- 1;
	temp_len = first_token_len;
	string_index = 0;
	/* searching protocol name from first token string */
	while ( temp_len != -1 )
	{
		if ( (first_token_string[temp_len] == ' ')
			|| (first_token_string[temp_len] == '\t')
			|| (first_token_string[temp_len] == '.')
		   )
		{
			temp_len--;
			continue;
		}
		else if (	(first_token_string[temp_len] == '&')
				||	(first_token_string[temp_len] == '!')
				||	(first_token_string[temp_len] == '(')
				||	(first_token_string[temp_len] == '|')
				)
					break;
		else
			protocol_name[string_index++] = first_token_string[temp_len--];
	}
	protocol_name[string_index] = '\0';
	/* checking for string before protocol name */
	if ( temp_len < 0 )
		first_string_prepend_flag = 0;

	temp_len = 0;
	/* reversing the string protocol name */
	strrev(protocol_name);

	/* checking for protocol name */
	if (	( 0 != strcmp(protocol_name,"api_call") )
		 && ( 0 != strcmp(protocol_name,"api_rc") )
		 && ( 0 != strcmp(protocol_name,"api_mn") )
		 && ( 0 != strcmp(protocol_name,"frx_if") )
		 && ( 0 != strcmp(protocol_name,"fap_if") )
		 && ( 0 != strcmp(protocol_name,"dct") )
		 && ( 0 != strcmp(protocol_name,"ipcs") )
		 && ( 0 != strcmp(protocol_name,"sicap") )
		 && ( 0 != strcmp(protocol_name,"api") )
		)
		goto CLEANUP;
	if ( 1 == first_string_prepend_flag )
	{
		temp_len = first_token_len - strlen(protocol_name);
		/* allocating memory for first string to be prepend */
		first_string_prepend = (char *) calloc ( temp_len + 1, sizeof (char) );
		if ( NULL == first_string_prepend )
			goto CLEANUP;
		/* copying first_token_string in first_string_prepend */
		strncpy(first_string_prepend,first_token_string,temp_len);
	}
	/* freeing memory for first_token_string */
	free(first_token_string);
	first_token_string = NULL;

	memset(parameter_name, 0, sizeof(parameter_name));
	memset(expr, 0, sizeof(expr));
	/* to start length and index both from index 0 */
	second_token_len = strlen(second_token_string)- 1;
	temp_len = second_token_len;
	string_index = 1;
	expr_index = 0;
	param_index = 0;
	expr_flag = 0;
	length_count = 1;
	/* searching parameter name and expression from second token string */
	while ( string_index <= temp_len )
	{
		if (	(second_token_string[string_index] == '\0' )
				||	(second_token_string[string_index] == ')')
				||	(second_token_string[string_index] == '&')
				||	(second_token_string[string_index] == '|')
			)
					break;
		else if (	(second_token_string[string_index] == ' ' )
				||	(second_token_string[string_index] == '\t' )
				)
		{
			length_count++;
			string_index++;
		}
		else if ( (second_token_string[string_index] == '=' )
				|| (second_token_string[string_index] == '!' )
				|| (second_token_string[string_index] == '<' )
				|| (second_token_string[string_index] == '>' )
				)
		{
			expr_flag = 1;
			expr[expr_index++] = second_token_string[string_index++];
			length_count++;
			if ( expr_index >= MAX_EXPRESSION_LENGTH )
				break;
		}
		else
		{
			if ( 1 == expr_flag )
				expr[expr_index++] = second_token_string[string_index++];
			else
				parameter_name[param_index++] = second_token_string[string_index++];
			length_count++;
			if ( expr_index >= MAX_EXPRESSION_LENGTH )
				break;
		}
	}
	parameter_name[param_index] = '\0';
	if ( expr_index >= MAX_EXPRESSION_LENGTH )
		expr[MAX_EXPRESSION_LENGTH-1] = '\0';
	else
		expr[expr_index] = '\0';

	if ( (NULL == parameter_name) || (NULL == protocol_name) )
		goto CLEANUP;

	/* checking for string after expr */
	if ( second_token_string[length_count] == '\0' )
			second_string_append_flag = 0;

	if ( 1 == second_string_append_flag )
	{
		temp_len = (second_token_len + 1) - length_count ;
		/* allocating memory for second string to be append */
		second_string_append = (char *) calloc ( temp_len + 1, sizeof (char) );
		if ( NULL == second_string_append )
			goto CLEANUP;
		/* copying second_token_string in second_string_append */
		strncpy(second_string_append,(second_token_string+length_count),temp_len);
	}
	/* freeing memory for second_token_string */
	free(second_token_string);
	second_token_string = NULL;

	token = NULL;

	if (0 == strcmp(protocol_name,"api"))
	{
		filter_string_apicall = create_filter("api_call", parameter_name, expr);
		filter_string_api_rc = create_filter("api_rc", parameter_name, expr);
		filter_string_api_mn = create_filter("api_mn", parameter_name, expr);

		if (filter_string_apicall)
		{
			filter_string =  calloc(strlen(filter_string_apicall)+1, sizeof(char));
			strncat(filter_string, filter_string_apicall, strlen(filter_string_apicall));
		}
		if (filter_string_api_rc)
		{
			if (filter_string)
			{
				len =strlen(filter_string)+strlen(filter_string_api_rc)+2+1;
				filter_string =  realloc(filter_string,
					len * sizeof (char));

				strncat(strncat(filter_string, "||", len), filter_string_api_rc, len);

			}
			else
			{
				filter_string =  calloc(strlen(filter_string_api_rc)+1, sizeof(char));
				strncat(filter_string, filter_string_api_rc, strlen(filter_string_api_rc));
			}
		}
		if (filter_string_api_mn)
		{
			if (filter_string)
			{
				len =strlen(filter_string)+strlen(filter_string_api_mn)+2+1;
				filter_string =  realloc(filter_string,
					len * sizeof (char));
				strncat(filter_string, "||", len);
				strncat(filter_string, filter_string_api_mn, len);

			}
			else
			{
				filter_string =  calloc(strlen(filter_string_api_mn)+1, sizeof(char));
				strncat(filter_string, filter_string_api_mn, strlen(filter_string_api_mn));
			}
		}


	}
	else
	{
		filter_string = create_filter(protocol_name, parameter_name, expr);
	}
	if ( NULL == filter_string )
		goto CLEANUP;

	/* concating first_string_prepend, filter_string and second_string_append */
	if ( (NULL != first_string_prepend ) && (NULL != second_string_append ) )
	{
		final_filter_string = (char *) calloc ( strlen(first_string_prepend)
											+ strlen(filter_string)
											+ strlen(second_string_append)
											+ 1,
											sizeof (char)
										 );
		if ( NULL == final_filter_string )
			goto CLEANUP;
		sprintf( final_filter_string, "%s%s%s",first_string_prepend
											  , filter_string
											  , second_string_append
				);
	}
	else if ( (NULL != first_string_prepend ) && (NULL == second_string_append ) )
	{
		final_filter_string = (char *) calloc ( strlen(first_string_prepend)
											+ strlen(filter_string)
											+ 1,
											sizeof (char)
										 );
		if ( NULL == final_filter_string )
			goto CLEANUP;
		sprintf( final_filter_string, "%s%s",first_string_prepend, filter_string);
	}
	else if ( (NULL == first_string_prepend ) && (NULL != second_string_append ) )
	{
		final_filter_string = (char *) calloc ( strlen(filter_string)
												+ strlen(second_string_append)
												+ 1,
												sizeof (char)
											  );
		if ( NULL == final_filter_string )
			goto CLEANUP;
		sprintf( final_filter_string, "%s%s",filter_string, second_string_append);
	}
	else
	{
		final_filter_string = (char *) calloc ( strlen(filter_string)
												+ 1,
												sizeof (char)
											  );
		if ( NULL == final_filter_string )
			goto CLEANUP;
		sprintf( final_filter_string, "%s",filter_string);
	}

	if ( NULL != temp_filter_string )
	{
		free(temp_filter_string);
		temp_filter_string = NULL;
	}
	if ( NULL != first_token_string )
	{
		free(first_token_string);
		first_token_string = NULL;
	}
	if ( NULL != second_token_string )
	{
		free(second_token_string);
		second_token_string = NULL;
	}
	if ( NULL != first_string_prepend )
	{
		free(first_string_prepend);
		first_string_prepend = NULL;
	}
	if ( NULL != second_string_append )
	{
		free(second_string_append);
		second_string_append = NULL;
	}
	if ( NULL != filter_string )
	{
		free(filter_string);
		filter_string = NULL;
	}

	if ( NULL != filter_string_apicall)
	{
		free(filter_string_apicall);
 		filter_string_apicall = NULL;
	}

	if ( NULL != filter_string_api_rc)
	{
		free(filter_string_api_rc);
		filter_string_api_rc = NULL;
	}

	if ( NULL != filter_string_api_mn)

	{
		free(filter_string_api_mn);
		filter_string_api_mn = NULL;
	}

	if ( NULL != final_filter_string )
		return final_filter_string;
	else
		return NULL;

CLEANUP:

	if ( NULL != temp_filter_string )
	{
		free(temp_filter_string);
		temp_filter_string = NULL;
	}
	if ( NULL != first_token_string )
	{
		free(first_token_string);
		first_token_string = NULL;
	}
	if ( NULL != second_token_string )
	{
		free(second_token_string);
		second_token_string = NULL;
	}
	if ( NULL != first_string_prepend )
	{
		free(first_string_prepend);
		first_string_prepend = NULL;
	}
	if ( NULL != second_string_append )
	{
		free(second_string_append);
		second_string_append = NULL;
	}
	if ( NULL != filter_string )
	{
		free(filter_string);
		filter_string = NULL;
	}
	if ( NULL != final_filter_string )
	{
		free(final_filter_string);
		final_filter_string = NULL;
	}

	if ( NULL != filter_string_apicall)
	{
		free(filter_string_apicall);
 		filter_string_apicall = NULL;
	}

	if ( NULL != filter_string_api_rc)
	{
		free(filter_string_api_rc);
		filter_string_api_rc = NULL;
	}

	if ( NULL != filter_string_api_mn)

	{
		free(filter_string_api_mn);
		filter_string_api_mn = NULL;
	}

	return NULL;
}



char*
replace_string( char *input_str,
				char *find_str,
				char *replace_str)
{
	const char *find_ptr;
	char *replace_ptr = NULL;
	char *result_str = NULL;
	unsigned int replace_counter;
	unsigned int find_counter;
	unsigned int find_str_len = 0;
	unsigned int replace_str_len = 0;
	int match_flag = 0;



	if ( NULL == input_str
		|| NULL == find_str
		|| NULL == replace_str
		)
		return input_str;

	find_str_len = strlen(find_str);
	replace_str_len = strlen(replace_str);

	result_str = (char*)calloc( (strlen(input_str)*replace_str_len/find_str_len)+1,
								sizeof(char) );

	if ( NULL == result_str )
		return input_str;

	for( find_ptr = input_str, replace_ptr = result_str; *find_ptr != '\0'; find_ptr++ )
	{
		/* logic to find the exact string */
		for( find_counter = 0; find_counter < find_str_len; find_counter++ )
		{
			if ( *(find_ptr+find_counter) == *(find_str+find_counter) )
				match_flag = 1;
			else
			{
				match_flag = 0;
				break;
			}
			if ( (find_counter == (find_str_len-1)) && (1 == match_flag) )
			{
				if ( (*(find_ptr+find_counter+1) != '\0')
					&& (*(find_ptr+find_counter+1) != ')')
					&& (*(find_ptr+find_counter+1) != '|')
					&& (*(find_ptr+find_counter+1) != '&')
					&& (*(find_ptr+find_counter+1) != ' ')
					&& (*(find_ptr+find_counter+1) != '\t')
					)
				match_flag = 0;
			}
		}
		/* logic to replace the string */
		if( match_flag )
		{
			for( replace_counter = 0; replace_counter < replace_str_len; replace_counter++ )
				*(replace_ptr+replace_counter) = *(replace_str+replace_counter);

			replace_ptr = replace_ptr+replace_counter;
			find_ptr = find_ptr+find_counter-1;
		}
		/* copying unchanged characters */
		else
		{
			*replace_ptr = *find_ptr;
			replace_ptr++;
		}
	}
	*replace_ptr = '\0';

	return result_str;
}
/* PMC Wildcard, hello, soap Support - end */

int
findString(char *path, char *api_id)
{
	FILE *config_fp;
	char config_line[MAX_LINE_LENGTH];
	char *api_id_match = NULL;

	if ( (path == NULL)|| (api_id ==  NULL) )
		return 0;

    config_fp = fopen(path, "r");
	if (config_fp == NULL)
		return 0;

	memset(config_line, 0, sizeof(config_line));
	/* reading lines from api_mn/api_call configuration file */
	while(fgets(config_line, sizeof(config_line), config_fp) != NULL)
	{
		/* checking for blank lines  */
		if(is_blank_line(config_line))
			continue;
		/* matching api_id in the line */
		api_id_match = strstr(config_line ,api_id);
		if (api_id_match != NULL)
		 {
			 fclose(config_fp);
			 return 1;
		 }
	}
	fclose(config_fp);
	return 0;
}

gboolean
dfilter_compile(const gchar *text, dfilter_t **dfp)
{
	int		token;
	dfilter_t	*dfilter;
	dfwork_t	*dfw;
	gboolean failure = FALSE;
	const char	*depr_test;
	guint		i;
	/* PMC Wildcard, hello, soap Support	-  start */

	char * temp_first_token_string = NULL;  /*
											** for filter string
											** e.g. api_call.*.TimerId
											** to check the presence of no. of
											** wildcard(*) in string. Now
											** only 1 wildcard is supported.
											*/
	char * hello_soap_filter_string = NULL; /*
											** for replacement of hello and
											** soap from filter string
											*/
	char * temp_hello_soap_string = NULL; /* temp hello soap string */
	char * hello_flag = NULL;/* flag for hello presence in filter string */
	char * soap_flag = NULL;/* flag for soap presence in filter string */
	char * api_flag = NULL;	/* flag for "api" string presence in filter string */
	char * hello = "hello";	/* hello string */
	char * replace_hello = "(mmtss.header.message_type==0x000c)";
							/*
							** replace hello string
							*/
	char * soap = "soap";	/* soap string */
	char * replace_soap = "(xml)"; /* replace soap string */
	char * api = "api";
	char * replace_api = "((mmtss.header.message_type==0x0006)||(mmtss.header.message_type==0x0007)||(mmtss.header.message_type==0x0008))";
	char * api_dot_string = NULL;
	char * api_dot_token = NULL;
	char * rep = NULL;
	char * api_id = NULL;
	char * replace_api_dot_string = NULL;
	gboolean mn_flag = FALSE;
	gboolean call_flag = FALSE;
	char * mn_config_file_name = NULL;
	char * call_config_file_name = NULL;
	char ** mn_config_file_path;
	char ** call_config_file_path;

	int count_asterik; /* count of wildcard(*) present in the filter string */
	int str_index = 0;		/* for filter string index */

	/* PMC Wildcard, hello, soap Support	- end */
	GPtrArray	*deprecated = g_ptr_array_new();

	dfilter_error_msg = NULL;

	if ( !( text = dfilter_macro_apply(text, 0, &dfilter_error_msg) ) ) {
		return FALSE;
	}

	dfw = dfwork_new();


/* PMC wildcard,hello,soap,api Support - start */

	/* df_scanner_text(text); */
	/* checking for hello flag in the filter string */
	hello_flag = strstr( text, "hello" );
	/* checking for soap flag in the filter string */
	soap_flag = strstr( text, "soap" );
	/* checking for "api" text in the filter string */
	api_flag = strstr( text, "api" );


#ifdef MMT_DCM
	if ( (NULL != hello_flag) || (NULL != soap_flag) )
		goto FAILURE;
#endif
	/*
	** finding and replacing "hello","soap","api"
	** from filter string.
	*/

	temp_hello_soap_string = (char *) calloc ( strlen (text)+1
		, sizeof (char)
		);
	if ( NULL == temp_hello_soap_string )
		goto FAILURE;

	/* copying original filter string in a temporary string */
	strncpy( temp_hello_soap_string, text, strlen(text) );

	/* assigning temp filter string in a hello-soap string */
	hello_soap_filter_string = temp_hello_soap_string;

	/* replacing "hello" string */
	if ( NULL != hello_flag )
		/* calling function to replace "hello" from filter string */
		hello_soap_filter_string = replace_string(
		hello_soap_filter_string
		,hello
		,replace_hello
		);

	/* replacing "soap" string */
	if ( NULL != soap_flag )
		/* calling function to replace "soap" from filter string */
		hello_soap_filter_string = replace_string(
		hello_soap_filter_string
		,soap
		,replace_soap
		);

	/* replacing "api" string */
	if ( NULL != api_flag )
		/* calling function to replace "api" from filter string */
		hello_soap_filter_string = replace_string(
		hello_soap_filter_string
		,api
		,replace_api
		);

	/*Support for API_DOT(api.1209) filter for CALL,MN,RC  start*/
		api_dot_string = (char *) calloc(strlen (hello_soap_filter_string)+1
								,sizeof (char)
								);
		if (api_dot_string == NULL)
			return FALSE;


		strcpy(api_dot_string ,hello_soap_filter_string);
		api_dot_token = strtok (api_dot_string, " &|()!");

		while (api_dot_token != NULL)
		{
		  if(((strncmp(api_dot_token, "api.*.", 6)!= 0)) && (strncmp(api_dot_token, "api.", 4)== 0))
		  {
			  /* Extracting api id  from string */
			  api_id = strstr(api_dot_token, ".");
			  /* Loading libwireshark.dll for absoulte path */
			  if (NULL == wiresharkLib)
			   {
				  wiresharkLib = LoadLibrary(TEXT("libwireshark.dll"));
				  if (NULL == wiresharkLib)
					  return NULL;
			   }
			  mn_config_file_path = (char**) GetProcAddress(wiresharkLib, "full_path_name");
			  /* Storing API_MN configuration file path */
			  mn_config_file_name = (char *) calloc(strlen (*mn_config_file_path)
													+ strlen("\\Config\\api_mn.txt")
													+ 1 ,sizeof (char)
													 );
			  if (mn_config_file_name == NULL)
				  return NULL;
			  sprintf(mn_config_file_name,"%s\\Config\\api_mn.txt" ,*mn_config_file_path);

			   /* Preparing filter string */
			  if (findString(mn_config_file_name, api_id) == 1)
				  mn_flag = TRUE;

              call_config_file_path = (char**) GetProcAddress(wiresharkLib, "full_path_name");
			  /* Storing API_MN configuration file path */
			  call_config_file_name = (char *) calloc(strlen (*call_config_file_path)
													+ strlen("\\Config\\api_call.txt")
													+ 1 ,sizeof (char)
													 );
			  if (call_config_file_name == NULL)
				  return NULL;
			  sprintf(call_config_file_name,"%s\\Config\\api_call.txt" ,*call_config_file_path);

			   /* Preparing filter string */
			  if (findString(call_config_file_name, api_id) == 1)
				  call_flag = TRUE;


			  replace_api_dot_string = (char *) calloc(strlen ("(api_call||api_rc||api_mn)")
		   											   +strlen(api_id)* 3 +1
													   ,sizeof (char)
														);
			  if (replace_api_dot_string == NULL)
				  return NULL;


			  if (mn_flag == TRUE && call_flag == TRUE)
			  {
				sprintf(replace_api_dot_string, "(api_rc%s||api_call%s||api_mn%s)",api_id,api_id,api_id);
				 mn_flag = FALSE;
				 call_flag = FALSE;
			  }
			  else if (mn_flag == TRUE)
			  {
			     sprintf(replace_api_dot_string, "(api_rc%sapi_mn%s)",api_id,api_id);
				 mn_flag = FALSE;
			  }
			  else if (call_flag == TRUE)
			  {
			     sprintf(replace_api_dot_string, "(api_rc%s||api_call%s)",api_id,api_id);
				 call_flag = FALSE;
			  }
			  else
				 sprintf(replace_api_dot_string, "(api_rc%s)",api_id);

			  hello_soap_filter_string = replace_string(
			  hello_soap_filter_string
			  ,api_dot_token
			  ,replace_api_dot_string);

			  /* freeing all memory*/
			  free(replace_api_dot_string);
			  replace_api_dot_string = NULL;

			  free(mn_config_file_name);
			  mn_config_file_name = NULL;
		  }
		  api_dot_token = strtok (NULL, " &|()!");
		}

		free(api_dot_string);
		api_dot_string = NULL;



	/*Support for API_DOT(api.1209) filter for CALL,MN,RC  End*/

	/*  replacing wildcard(*) from filter string */
	if ( NULL != strchr(hello_soap_filter_string, '*') )
	{
		temp_first_token_string = (char *) calloc ( strlen (hello_soap_filter_string)+1
														, sizeof (char));
		if ( NULL == temp_first_token_string )
			goto FAILURE;
		/* copying original filter string in a temporary string */
		strncpy( temp_first_token_string, hello_soap_filter_string, strlen(hello_soap_filter_string) );

		count_asterik = 0;
		while ( temp_first_token_string[str_index] != '\0' )
		{
			if ( temp_first_token_string[str_index] == '\n' )
				goto FAILURE;
			/* checking for no. of wildcard(*) present in filter string */
			if ( temp_first_token_string[str_index] == '*' )
			{
				/*
				** checking for the format ".*."  for first
				** wildcard(*) in filter string
				*/
				if (   ( temp_first_token_string[str_index] == '*' )
					&& ( temp_first_token_string[str_index-1] == '.' )
					&& ( temp_first_token_string[str_index+1] != '\0')
					&& ( temp_first_token_string[str_index+1] == '.' )
					)
				{
					count_asterik++;
				}
			}
			str_index++;
		}
		free(temp_first_token_string);
		temp_first_token_string = NULL;

		while ( 0 != count_asterik )
		{
			hello_soap_filter_string = create_filter_string(hello_soap_filter_string);
			count_asterik --;
		}
	}

	if ( NULL == hello_soap_filter_string )
	{
		df_scanner_text(text);
	}
	else
	{
		df_scanner_text(hello_soap_filter_string);
	}

/* PMC wildcard,hello,soap,api Support - end */

	while (1) {
		df_lval = stnode_new(STTYPE_UNINITIALIZED, NULL);
		token = df_lex();

		/* Check for scanner failure */
		if (token == SCAN_FAILED) {
			failure = TRUE;
			break;
		}

		/* Check for end-of-input */
		if (token == 0) {
			break;
		}

		/* See if the node is deprecated */
		depr_test = stnode_deprecated(df_lval);

		if (depr_test) {
			for (i = 0; i < deprecated->len; i++) {
				if (strcasecmp(depr_test, g_ptr_array_index(deprecated, i)) == 0) {
					/* It's already in our list */
					depr_test = NULL;
				}
			}
		}

		if (depr_test) {
			g_ptr_array_add(deprecated, g_strdup(depr_test));
		}

		/* Give the token to the parser */
		Dfilter(ParserObj, token, df_lval, dfw);
		/* We've used the stnode_t, so we don't want to free it */
		df_lval = NULL;

		if (dfw->syntax_error) {
			failure = TRUE;
			break;
		}

	} /* while (1) */

	/* If we created an stnode_t but didn't use it, free it; the
	 * parser doesn't know about it and won't free it for us. */
	if (df_lval) {
		stnode_free(df_lval);
		df_lval = NULL;
	}

	/* Tell the parser that we have reached the end of input; that
	 * way, it'll reset its state for the next compile.  (We want
	 * to do that even if we got a syntax error, to make sure the
	 * parser state is cleaned up; we don't create a new parser
	 * object when we start a new parse, and don't destroy it when
	 * the parse finishes.) */
	Dfilter(ParserObj, 0, NULL, dfw);

	/* One last check for syntax error (after EOF) */
	if (dfw->syntax_error)
		failure = TRUE;

	/* Reset flex */
	df_scanner_cleanup();

	if (failure)
		goto FAILURE;

	/* Success, but was it an empty filter? If so, discard
	 * it and set *dfp to NULL */
	if (dfw->st_root == NULL) {
		*dfp = NULL;
		for (i = 0; i < deprecated->len; ++i) {
			gchar* depr = g_ptr_array_index(deprecated,i);
			g_free(depr);
		}
		g_ptr_array_free(deprecated, TRUE);
	}
	else {

		/* Check semantics and do necessary type conversion*/
		if (!dfw_semcheck(dfw)) {
			goto FAILURE;
		}

		/* Create bytecode */
		dfw_gencode(dfw);

		/* Tuck away the bytecode in the dfilter_t */
		dfilter = dfilter_new();
		dfilter->insns = dfw->insns;
		dfilter->consts = dfw->consts;
		dfw->insns = NULL;
		dfw->consts = NULL;
		dfilter->interesting_fields = dfw_interesting_fields(dfw,
			&dfilter->num_interesting_fields);

		/* Initialize run-time space */
		dfilter->num_registers = dfw->first_constant;
		dfilter->max_registers = dfw->next_register;
		dfilter->registers = g_new0(GList*, dfilter->max_registers);
		dfilter->attempted_load = g_new0(gboolean, dfilter->max_registers);

		/* Initialize constants */
		dfvm_init_const(dfilter);

		/* Add any deprecated items */
		dfilter->deprecated = deprecated;

		/* And give it to the user. */
		*dfp = dfilter;
	}
	/* SUCCESS */
	dfwork_free(dfw);
	/* PMC wildcard Support - start */
	if ( NULL != temp_hello_soap_string )
	{
		free(temp_hello_soap_string);
		temp_hello_soap_string = NULL;
	}
	/* PMC wildcard Support	- end */
	return TRUE;

FAILURE:
	if (dfw) {
		dfwork_free(dfw);
	}
	for (i = 0; i < deprecated->len; ++i) {
		gchar* depr = g_ptr_array_index(deprecated,i);
		g_free(depr);
	}
	g_ptr_array_free(deprecated, TRUE);
	dfilter_fail("Unable to parse filter string \"%s\".", text);

	*dfp = NULL;
	/* PMC wildcard Support - start */
	if ( NULL != temp_hello_soap_string )
	{
		free(temp_hello_soap_string);
		temp_hello_soap_string = NULL;
	}
	/* PMC wildcard Support	- end */
	return FALSE;
}


gboolean
dfilter_apply(dfilter_t *df, proto_tree *tree)
{
	return dfvm_apply(df, tree);
}

gboolean
dfilter_apply_edt(dfilter_t *df, epan_dissect_t* edt)
{
	return dfvm_apply(df, edt->tree);
}


void
dfilter_prime_proto_tree(const dfilter_t *df, proto_tree *tree)
{
    int i;

    for (i = 0; i < df->num_interesting_fields; i++) {
        proto_tree_prime_hfid(tree, df->interesting_fields[i]);
    }
}

GPtrArray *
dfilter_deprecated_tokens(dfilter_t *df) {
	if (df->deprecated && df->deprecated->len > 0) {
		return df->deprecated;
	}
	return NULL;
}

void
dfilter_dump(dfilter_t *df)
{
	guint i;
	gchar *sep = "";

	dfvm_dump(stdout, df->insns);

	if (df->deprecated && df->deprecated->len) {
		printf("\nDeprecated tokens: ");
		for (i = 0; i < df->deprecated->len; i++) {
			printf("%s\"%s\"", sep, (char *) g_ptr_array_index(df->deprecated, i));
			sep = ", ";
		}
		printf("\n");
	}
}
